package gr.uom.java.xmi;

import java.io.File;

import java.io.IOException;

import java.util.LinkedHashSet;

import java.util.List;

import org.apache.commons.io.FileUtils;

import org.eclipse.jdt.core.dom.AST;

import org.eclipse.jdt.core.dom.ASTParser;

import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;

import org.eclipse.jdt.core.dom.CompilationUnit;

import org.eclipse.jdt.core.dom.FieldDeclaration;

import org.eclipse.jdt.core.dom.MethodDeclaration;

import org.eclipse.jdt.core.dom.Modifier;

import org.eclipse.jdt.core.dom.PackageDeclaration;

import org.eclipse.jdt.core.dom.SingleVariableDeclaration;

import org.eclipse.jdt.core.dom.Type;

import org.eclipse.jdt.core.dom.TypeDeclaration;

import org.eclipse.jdt.core.dom.VariableDeclarationFragment;

public class ASTReader {

	private UMLModel umlModel;

	private UMLModel bytecodeModel;
	public static long elementID = 0;

	public ASTReader(File rootFile) {

		this.umlModel = new UMLModel();

		this.bytecodeModel = new BytecodeReader(rootFile).getUmlModel();

		recurse(rootFile);

	}

	public UMLModel getUmlModel() {

		return umlModel;

	}

	private void recurse(File rootFile) {

		if (rootFile.isDirectory()) {

			File[] files = rootFile.listFiles();

			for (File file : files) {

				if (file.isDirectory())

					recurse(file);

				else {

					String fileName = file.getName();

					if (fileName.contains(".")) {

						String extension = fileName.substring(fileName
								.lastIndexOf("."));

						if (extension.equalsIgnoreCase(".java")) {

							parseAST(file);

						}

					}

				}

			}

		}

		else {

			String fileName = rootFile.getName();

			if (fileName.contains(".")) {

				String extension = fileName
						.substring(fileName.lastIndexOf("."));

				if (extension.equalsIgnoreCase(".java")) {

					parseAST(rootFile);

				}

			}

		}

	}

	private void parseAST(File javaFile) {

		try {

			String source = FileUtils.readFileToString(javaFile);

			ASTParser parser = ASTParser.newParser(AST.JLS3);

			parser.setKind(ASTParser.K_COMPILATION_UNIT);

			parser.setSource(source.toCharArray());

			// parser.setResolveBindings(true);

			CompilationUnit compilationUnit = (CompilationUnit) parser
					.createAST(null);

			PackageDeclaration packageDeclaration = compilationUnit
					.getPackage();

			String packageName = null;

			if (packageDeclaration != null)

				packageName = packageDeclaration.getName()
						.getFullyQualifiedName();

			else

				packageName = "";

			List<AbstractTypeDeclaration> topLevelTypeDeclarations = compilationUnit
					.types();

			for (AbstractTypeDeclaration abstractTypeDeclaration : topLevelTypeDeclarations) {

				if (abstractTypeDeclaration instanceof TypeDeclaration) {

					TypeDeclaration topLevelTypeDeclaration = (TypeDeclaration) abstractTypeDeclaration;

					processTypeDeclaration(topLevelTypeDeclaration, packageName);

				}

			}

		} catch (IOException e) {

			e.printStackTrace();

		}

	}

	private void processTypeDeclaration(TypeDeclaration typeDeclaration,
			String packageName) {

		String className = typeDeclaration.getName().getFullyQualifiedName();

		UMLClass umlClass = new UMLClass(packageName, className, null,
				ASTReader.elementID++);

		UMLClass bytecodeClass = bytecodeModel.getClass(umlClass.getName());

		if (typeDeclaration.isInterface()) {

			umlClass.setInterface(true);

		}

		int modifiers = typeDeclaration.getModifiers();

		if ((modifiers & Modifier.ABSTRACT) != 0)

			umlClass.setAbstract(true);

		if ((modifiers & Modifier.PUBLIC) != 0)

			umlClass.setVisibility("public");

		else if ((modifiers & Modifier.PROTECTED) != 0)

			umlClass.setVisibility("protected");

		else if ((modifiers & Modifier.PRIVATE) != 0)

			umlClass.setVisibility("private");

		else

			umlClass.setVisibility("package");

		Type superclassType = typeDeclaration.getSuperclassType();

		if (superclassType != null) {

			UMLGeneralization umlGeneralization = new UMLGeneralization(
					umlClass.getName(), superclassType.toString());

			UMLGeneralization bytecodeGeneralization = bytecodeModel
					.matchGeneralization(umlGeneralization);

			if (bytecodeGeneralization != null) {

				umlGeneralization.setParent(bytecodeGeneralization.getParent());

			}

			umlModel.addGeneralization(umlGeneralization);

		}

		List<Type> superInterfaceTypes = typeDeclaration.superInterfaceTypes();

		for (Type interfaceType : superInterfaceTypes) {

			UMLRealization umlRealization = new UMLRealization(umlClass
					.getName(), interfaceType.toString());

			UMLRealization bytecodeRealization = bytecodeModel
					.matchRealization(umlRealization);

			if (bytecodeRealization != null) {

				umlRealization.setSupplier(bytecodeRealization.getSupplier());

			}

			umlModel.addRealization(umlRealization);

		}

		FieldDeclaration[] fieldDeclarations = typeDeclaration.getFields();

		for (FieldDeclaration fieldDeclaration : fieldDeclarations) {

			Type fieldType = fieldDeclaration.getType();

			List<VariableDeclarationFragment> fragments = fieldDeclaration
					.fragments();

			for (VariableDeclarationFragment fragment : fragments) {

				String fieldName = fragment.getName().getFullyQualifiedName();
				int extraDimensions = fragment.getExtraDimensions();
				String fieldTypeString = fieldType.toString();
				for (int i = 0; i < extraDimensions; i++) {
					fieldTypeString += "[]";
				}

				UMLAttribute umlAttribute = new UMLAttribute(fieldName,
						fieldTypeString, ASTReader.elementID++);

				int fieldModifiers = fieldDeclaration.getModifiers();

				if ((fieldModifiers & Modifier.PUBLIC) != 0)

					umlAttribute.setVisibility("public");

				else if ((fieldModifiers & Modifier.PROTECTED) != 0)

					umlAttribute.setVisibility("protected");

				else if ((fieldModifiers & Modifier.PRIVATE) != 0)

					umlAttribute.setVisibility("private");

				else

					umlAttribute.setVisibility("package");

				if (bytecodeClass != null) {

					UMLAttribute bytecodeAttribute = bytecodeClass
							.matchAttribute(umlAttribute);

					if (bytecodeAttribute != null)

						umlAttribute.setType(bytecodeAttribute.getType());

				}

				umlClass.addAttribute(umlAttribute);

			}

		}

		MethodDeclaration[] methodDeclarations = typeDeclaration.getMethods();

		for (MethodDeclaration methodDeclaration : methodDeclarations) {

			String methodName = methodDeclaration.getName()
					.getFullyQualifiedName();

			UMLOperation umlOperation = new UMLOperation(methodName, null,
					ASTReader.elementID++);

			umlOperation.setClassName(umlClass.getName());

			int methodModifiers = methodDeclaration.getModifiers();

			if ((methodModifiers & Modifier.PUBLIC) != 0)

				umlOperation.setVisibility("public");

			else if ((methodModifiers & Modifier.PROTECTED) != 0)

				umlOperation.setVisibility("protected");

			else if ((methodModifiers & Modifier.PRIVATE) != 0)

				umlOperation.setVisibility("private");

			else

				umlOperation.setVisibility("package");

			if ((methodModifiers & Modifier.ABSTRACT) != 0)

				umlOperation.setAbstract(true);

			Type returnType = methodDeclaration.getReturnType2();

			if (returnType != null) {

				UMLParameter returnParameter = new UMLParameter("return",
						returnType.toString(), "return", ASTReader.elementID++);

				umlOperation.addParameter(returnParameter);

			}

			List<SingleVariableDeclaration> parameters = methodDeclaration
					.parameters();

			for (SingleVariableDeclaration parameter : parameters) {

				Type parameterType = parameter.getType();
				int extraDimensions = parameter.getExtraDimensions();
				String parameterTypeString = parameterType.toString();
				for (int i = 0; i < extraDimensions; i++) {
					parameterTypeString += "[]";
				}

				String parameterName = parameter.getName()
						.getFullyQualifiedName();

				UMLParameter umlParameter = new UMLParameter(parameterName,
						parameterTypeString, "in", ASTReader.elementID++);

				umlOperation.addParameter(umlParameter);

			}

			if (bytecodeClass != null) {

				UMLOperation bytecodeOperation = bytecodeClass
						.matchOperation(umlOperation);

				if (bytecodeOperation != null) {

					int i = 0;

					for (UMLParameter bytecodeParameter : bytecodeOperation
							.getParameters()) {

						umlOperation.getParameters().get(i).setType(
								bytecodeParameter.getType());

						i++;

					}

					umlOperation
							.setAccessedMembers(new LinkedHashSet<AccessedMember>(
									bytecodeOperation.getAccessedMembers()));

				}

			}

			umlClass.addOperation(umlOperation);

		}

		umlModel.addClass(umlClass);

		TypeDeclaration[] types = typeDeclaration.getTypes();

		for (TypeDeclaration type : types) {

			processTypeDeclaration(type, packageName + "." + className);

		}

	}

}
