package gr.uom.java.xmi;



import java.io.DataInputStream;

import java.io.File;

import java.io.FileInputStream;

import java.io.FileNotFoundException;

import java.io.IOException;

import java.util.Iterator;

import java.util.List;



import org.objectweb.asm.ClassReader;

import org.objectweb.asm.Opcodes;

import org.objectweb.asm.Type;

import org.objectweb.asm.tree.AbstractInsnNode;

import org.objectweb.asm.tree.ClassNode;

import org.objectweb.asm.tree.FieldInsnNode;

import org.objectweb.asm.tree.FieldNode;

import org.objectweb.asm.tree.MethodInsnNode;

import org.objectweb.asm.tree.MethodNode;

import org.objectweb.asm.tree.TypeInsnNode;



public class BytecodeReader {

	private UMLModel umlModel;



	public BytecodeReader(File rootFile) {

		this.umlModel = new UMLModel();

		recurse(rootFile);

	}



	public UMLModel getUmlModel() {

		return umlModel;

	}



	private void recurse(File rootFile) {

		if(rootFile.isDirectory()) {

			File[] files = rootFile.listFiles();

			for(File file : files) {

				if(file.isDirectory())

					recurse(file);

				else {

					String fileName = file.getName();

					if(fileName.contains(".")) {

						String extension = fileName.substring(fileName.lastIndexOf("."));

						if(extension.equalsIgnoreCase(".class")) {

							parseBytecode(file);

						}

					}

				}

			}

		}

		else {

			String fileName = rootFile.getName();

			if(fileName.contains(".")) {

				String extension = fileName.substring(fileName.lastIndexOf("."));

				if(extension.equalsIgnoreCase(".class")) {

					parseBytecode(rootFile);

				}

			}

		}

	}



	private void parseBytecode(File classFile) {

		try {

			FileInputStream fin = new FileInputStream(classFile);

			ClassReader cr = new ClassReader(new DataInputStream(fin));

			ClassNode cn = new ClassNode();

			cr.accept(cn, ClassReader.SKIP_DEBUG);



			String qualifiedClassName = cn.name.replaceAll("/", ".").replaceAll("\\$", ".");

			String packageName = null;

			String className = null;

			if(qualifiedClassName.contains(".")) {

				packageName = qualifiedClassName.substring(0, qualifiedClassName.lastIndexOf("."));

				className = qualifiedClassName.substring(qualifiedClassName.lastIndexOf(".")+1);

			}

			else {

				packageName = "";

				className = qualifiedClassName;

			}



			UMLClass umlClass = new UMLClass(packageName, className, null, ASTReader.elementID++);

			if((cn.access & Opcodes.ACC_INTERFACE) != 0)

				umlClass.setInterface(true);

			else if((cn.access & Opcodes.ACC_ABSTRACT) != 0)

				umlClass.setAbstract(true);



			if((cn.access & Opcodes.ACC_PUBLIC) != 0)

				umlClass.setVisibility("public");

			else if((cn.access & Opcodes.ACC_PROTECTED) != 0)

				umlClass.setVisibility("protected");

			else if((cn.access & Opcodes.ACC_PRIVATE) != 0)

				umlClass.setVisibility("private");

			else

				umlClass.setVisibility("package");



			//if ((cn.access & Opcodes.ACC_STATIC) != 0)

			//	umlClass.setStatic(true);

			String superClassType = cn.superName;

			if(!superClassType.equals("java/lang/Object")) {

				UMLGeneralization umlGeneralization = new UMLGeneralization(umlClass.getName(), superClassType.replaceAll("/", ".").replaceAll("\\$", "."));

				umlModel.addGeneralization(umlGeneralization);

			}



			List<String> superInterfaceTypes = cn.interfaces;

			for(String interfaceType : superInterfaceTypes) {

				UMLRealization umlRealization = new UMLRealization(umlClass.getName(), interfaceType.replaceAll("/", ".").replaceAll("\\$", "."));

				umlModel.addRealization(umlRealization);

			}



			List<FieldNode> fields = cn.fields;

			for(FieldNode fieldNode : fields) {

				Type fieldType = Type.getType(fieldNode.desc);

				/*if(fieldNode.signature != null) {

					TraceSignatureVisitor v = new TraceSignatureVisitor(ClassReader.SKIP_DEBUG);

					SignatureReader r = new SignatureReader(fieldNode.signature);

					r.accept(v);

					String declaration = v.getDeclaration();

				}*/

				UMLAttribute umlAttribute = new UMLAttribute(fieldNode.name, fieldType.getClassName().replaceAll("\\$", "."), ASTReader.elementID++);



				if((fieldNode.access & Opcodes.ACC_PUBLIC) != 0)

					umlAttribute.setVisibility("public");

				else if((fieldNode.access & Opcodes.ACC_PROTECTED) != 0)

					umlAttribute.setVisibility("protected");

				else if((fieldNode.access & Opcodes.ACC_PRIVATE) != 0)

					umlAttribute.setVisibility("private");

				else

					umlAttribute.setVisibility("package");



				//if ((fieldNode.access & Opcodes.ACC_STATIC) != 0)

				//	umlAttribute.setStatic(true);

				umlClass.addAttribute(umlAttribute);

			}



			List<MethodNode> methods = cn.methods;

			for(MethodNode methodNode : methods) {

				String methodName = null;

				if(methodNode.name.equals("<init>"))

					methodName = className;

				else

					methodName = methodNode.name;

				UMLOperation umlOperation = new UMLOperation(methodName, null, ASTReader.elementID++);

				umlOperation.setClassName(umlClass.getName());



				if ((methodNode.access & Opcodes.ACC_PUBLIC) != 0)

					umlOperation.setVisibility("public");

				else if ((methodNode.access & Opcodes.ACC_PROTECTED) != 0)

					umlOperation.setVisibility("protected");

				else if ((methodNode.access & Opcodes.ACC_PRIVATE) != 0)

					umlOperation.setVisibility("private");

				else

					umlOperation.setVisibility("package");



				if(!methodNode.name.equals("<init>")) {

					Type returnType = Type.getReturnType(methodNode.desc);

					UMLParameter returnParameter = new UMLParameter("return", returnType.getClassName().replaceAll("\\$", "."), "return", ASTReader.elementID++);

					umlOperation.addParameter(returnParameter);



					if ((methodNode.access & Opcodes.ACC_ABSTRACT) != 0)

						umlOperation.setAbstract(true);

					//if ((methodNode.access & Opcodes.ACC_STATIC) != 0)

					//	umlOperation.setStatic(true);

				}



				if(methodNode.signature != null) {

					/*TraceSignatureVisitor v = new TraceSignatureVisitor(ClassReader.SKIP_DEBUG);

					SignatureReader r = new SignatureReader(methodNode.signature);

					r.accept(v);

					String declaration = v.getDeclaration();*/

				}

				else {

					Type[] argumentTypes = Type.getArgumentTypes(methodNode.desc);

					for(Type argumentType : argumentTypes) {

						UMLParameter umlParameter = new UMLParameter("", argumentType.getClassName().replaceAll("\\$", "."), "in", ASTReader.elementID++);

						boolean skipArgument = false;

						if(cn.name.contains("$") && methodNode.name.equals("<init>") && argumentType.getClassName().replaceAll("\\$", ".").equals(packageName))

							skipArgument = true;

						if(!skipArgument)

							umlOperation.addParameter(umlParameter);

					}

				}



				if (methodNode.instructions.size() > 0) {

					Iterator instructionIterator = methodNode.instructions.iterator();

					while(instructionIterator.hasNext()) {

						AbstractInsnNode instruction = (AbstractInsnNode)instructionIterator.next();

						if(instruction instanceof FieldInsnNode) {

							FieldInsnNode fieldInstruction = (FieldInsnNode) instruction;

							Type fieldType = Type.getType(fieldInstruction.desc);

							FieldAccess fieldAccess = new FieldAccess(fieldInstruction.owner.replaceAll("/", ".").replaceAll("\\$", "."),

									fieldType.getClassName(), fieldInstruction.name, ASTReader.elementID++);

							umlOperation.addAccessedMember(fieldAccess);

						}



						if ((instruction.getOpcode() == Opcodes.INVOKEVIRTUAL) ||

								(instruction.getOpcode() == Opcodes.INVOKESTATIC) ||

								(instruction.getOpcode() == Opcodes.INVOKESPECIAL) ||

								(instruction.getOpcode() == Opcodes.INVOKEINTERFACE)) {



							MethodInsnNode methodInstruction = (MethodInsnNode) instruction;

							MethodCall methodCall = new MethodCall(

									methodInstruction.owner.replaceAll("/", ".").replaceAll("\\$", "."), methodInstruction.name,

									Type.getReturnType(methodInstruction.desc).getClassName(), ASTReader.elementID++);

							Type[] argTypes = Type.getArgumentTypes(methodInstruction.desc);

							for (Type argType : argTypes)

								methodCall.addParameter(argType.getClassName());

							boolean isObjectConstructorCall = false;

							if(methodInstruction.name.equals("<init>")) {

								methodCall.setConstructorCall(true);

								if(methodCall.getOriginClassName().equals("java.lang.Object"))

									isObjectConstructorCall = true;

							}

							

							if(!isObjectConstructorCall)

								umlOperation.addAccessedMember(methodCall);

						}



						if ((instruction.getOpcode() == Opcodes.NEW) ||

								(instruction.getOpcode() == Opcodes.ANEWARRAY)) {

							TypeInsnNode typeInstruction = (TypeInsnNode) instruction;

						}

					}

				}



				umlClass.addOperation(umlOperation);

			}



			umlModel.addClass(umlClass);


			fin.close();
		} catch(FileNotFoundException fnfe) {

			fnfe.printStackTrace();

		} catch(IOException ioe) {

			ioe.printStackTrace();

		}

	}

}

