package gr.uom.java.xmi.diff;

import gr.uom.java.xmi.MethodCall;
import gr.uom.java.xmi.UMLAttribute;
import gr.uom.java.xmi.UMLOperation;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

public class UMLClassDiff implements Comparable<UMLClassDiff> {
	private String className;
	private List<UMLOperation> addedOperations;
	private List<UMLOperation> removedOperations;
	private List<UMLAttribute> addedAttributes;
	private List<UMLAttribute> removedAttributes;
	private Map<UMLAttribute, UMLAttribute> changedAttributes;
	private Map<UMLOperation, UMLOperation> changedOperationsNoParam;
	private List<UMLAttributeDiff> attributeDiffList;
	private List<UMLOperationDiff> operationDiffList;
	private List<UMLOperationBodyDiff> operationBodyDiffList;
	private boolean visibilityChanged;
	private String oldVisibility;
	private String newVisibility;
	private boolean abstractionChanged;
	private boolean oldAbstraction;
	private boolean newAbstraction;
	private Map<UMLOperation,String> changedOperations;
	
	public UMLClassDiff(String className) {
		this.className = className;
		this.addedOperations = new ArrayList<UMLOperation>();
		this.removedOperations = new ArrayList<UMLOperation>();
		this.addedAttributes = new ArrayList<UMLAttribute>();
		this.removedAttributes = new ArrayList<UMLAttribute>();
		this.attributeDiffList = new ArrayList<UMLAttributeDiff>();
		this.operationDiffList = new ArrayList<UMLOperationDiff>();
		this.operationBodyDiffList = new ArrayList<UMLOperationBodyDiff>();
		this.visibilityChanged = false;
		this.abstractionChanged = false;
		this.changedOperations = new LinkedHashMap<UMLOperation,String>();
		this.changedAttributes = new LinkedHashMap<UMLAttribute, UMLAttribute>();
		this.changedOperationsNoParam = new LinkedHashMap<UMLOperation, UMLOperation>();
	}

	public String getClassName() {
		return className;
	}

	public void reportAddedOperation(UMLOperation umlOperation) {
		this.addedOperations.add(umlOperation);
	}
	public void reportChangedOperation(UMLOperation operation, String param, int change){
		if(!changedOperations.containsKey(operation)){
			String str = "parameter: " + param + " has been ";
			if (change == 1)
				str+="added";
			else if (change == 2)
				str+= "deleted";
			else str+="changed";
			changedOperations.put(operation, str);
		}	
	}
	
	public void addOperationBodyDiff(UMLOperationBodyDiff operationBodyDiff) {
		this.operationBodyDiffList.add(operationBodyDiff);
	}

	public List<UMLOperationBodyDiff> getOperationBodyDiffList() {
		return operationBodyDiffList;
	}

	public List<MethodCall> getAddedMethodCalls() {
		List<MethodCall> addedMethodCalls = new ArrayList<MethodCall>();
		for(UMLOperationBodyDiff operationBodyDiff : operationBodyDiffList) {
			addedMethodCalls.addAll(operationBodyDiff.getAddedMethodCalls());
		}
		return addedMethodCalls;
	}
	
	public void reportChangedAttributes(UMLAttribute attr1, UMLAttribute attr2){
		this.changedAttributes.put(attr1,attr2);
	}
	
	public void reportChangedOperations(UMLOperation op1, UMLOperation op2){
		this.changedOperationsNoParam.put(op1,op2);
	}

	public void reportRemovedOperation(UMLOperation umlOperation) {
		this.removedOperations.add(umlOperation);
	}

	public void reportAddedAttribute(UMLAttribute umlAttribute) {
		this.addedAttributes.add(umlAttribute);
	}

	public void reportRemovedAttribute(UMLAttribute umlAttribute) {
		this.removedAttributes.add(umlAttribute);
	}

	public void setVisibilityChanged(boolean visibilityChanged) {
		this.visibilityChanged = visibilityChanged;
	}

	public void setOldVisibility(String oldVisibility) {
		this.oldVisibility = oldVisibility;
	}

	public void setNewVisibility(String newVisibility) {
		this.newVisibility = newVisibility;
	}

	public void setAbstractionChanged(boolean abstractionChanged) {
		this.abstractionChanged = abstractionChanged;
	}

	public void setOldAbstraction(boolean oldAbstraction) {
		this.oldAbstraction = oldAbstraction;
	}

	public void setNewAbstraction(boolean newAbstraction) {
		this.newAbstraction = newAbstraction;
	}

	public boolean isEmpty() {
		return addedOperations.isEmpty() && removedOperations.isEmpty() &&
			addedAttributes.isEmpty() && removedAttributes.isEmpty() &&
			operationDiffList.isEmpty() && attributeDiffList.isEmpty() &&
			changedOperations.isEmpty() && changedAttributes.isEmpty() &&
			changedOperationsNoParam.isEmpty() &&
			!visibilityChanged && !abstractionChanged;
	}

	public List<UMLOperation> getAddedOperations() {
		return addedOperations;
	}

	public List<UMLOperation> getRemovedOperations() {
		return removedOperations;
	}

	public List<UMLAttribute> getAddedAttributes() {
		return addedAttributes;
	}

	public List<UMLAttribute> getRemovedAttributes() {
		return removedAttributes;
	}

	public String getOldVisibility() {
		return oldVisibility;
	}

	public String getNewVisibility() {
		return newVisibility;
	}

	public boolean isOldAbstraction() {
		return oldAbstraction;
	}

	public boolean isNewAbstraction() {
		return newAbstraction;
	}

	public void checkForAttributeChanges() {
		Set<UMLAttribute> removedAttributesToBeRemoved = new LinkedHashSet<UMLAttribute>();
		Set<UMLAttribute> addedAttributesToBeRemoved = new LinkedHashSet<UMLAttribute>();
		for(UMLAttribute removedAttribute : removedAttributes) {
			for(UMLAttribute addedAttribute : addedAttributes) {
				if(removedAttribute.getName().equals(addedAttribute.getName())) {
					UMLAttributeDiff attributeDiff = new UMLAttributeDiff(removedAttribute, addedAttribute);
					addedAttributesToBeRemoved.add(addedAttribute);
					removedAttributesToBeRemoved.add(removedAttribute);
					attributeDiffList.add(attributeDiff);
					break;
				}
			}
		}
		this.addedAttributes.removeAll(addedAttributesToBeRemoved);
		this.removedAttributes.removeAll(removedAttributesToBeRemoved);
	}

	public void checkForOperationSignatureChanges() {
		Set<UMLOperation> removedOperationsToBeRemoved = new LinkedHashSet<UMLOperation>();
		Set<UMLOperation> addedOperationsToBeRemoved = new LinkedHashSet<UMLOperation>();
		for(UMLOperation removedOperation : removedOperations) {
			List<UMLOperation> matchingAddedOperations = new ArrayList<UMLOperation>();
			for(UMLOperation addedOperation : addedOperations) {
				if(removedOperation.getName().equals(addedOperation.getName())) {
					matchingAddedOperations.add(addedOperation);
				}
			}
			if(matchingAddedOperations.size() == 1) {
				UMLOperation addedOperation = matchingAddedOperations.get(0);
				UMLOperationDiff operationDiff = new UMLOperationDiff(removedOperation, addedOperation);
				addedOperationsToBeRemoved.add(addedOperation);
				removedOperationsToBeRemoved.add(removedOperation);
				operationDiffList.add(operationDiff);
			}
			else if(matchingAddedOperations.size() > 1) {
				UMLOperation addedOperation = null;
				for(UMLOperation operation : matchingAddedOperations) {
					if(operation.getParameters().size() == removedOperation.getParameters().size()) {
						addedOperation = operation;
						break;
					}
				}
				if(addedOperation != null) {
					UMLOperationDiff operationDiff = new UMLOperationDiff(removedOperation, addedOperation);
					addedOperationsToBeRemoved.add(addedOperation);
					removedOperationsToBeRemoved.add(removedOperation);
					operationDiffList.add(operationDiff);
				}
			}
		}
		this.addedOperations.removeAll(addedOperationsToBeRemoved);
		this.removedOperations.removeAll(removedOperationsToBeRemoved);
	}

	
	public Map<UMLAttribute, UMLAttribute> getChangedAttributes() {
		return changedAttributes;
	}

	public Map<UMLOperation, UMLOperation> getChangedOperations() {
		return changedOperationsNoParam;
	}

	public boolean isVisibilityChanged() {
		return visibilityChanged;
	}

	public boolean isAbstractionChanged() {
		return abstractionChanged;
	}

	public String toString() {
		StringBuilder sb = new StringBuilder();
		//if(!isEmpty())
			//sb.append(className).append(":").append("\n");
		if(visibilityChanged) {
			sb.append("\t").append("visibility changed from " + oldVisibility + " to " + newVisibility).append("\n");
		}
		if(abstractionChanged) {
			sb.append("\t").append("abstraction changed from " + (oldAbstraction ? "abstract" : "concrete") + " to " +
					(newAbstraction ? "abstract" : "concrete")).append("\n");
		}
		Collections.sort(removedOperations);
		for(UMLOperation umlOperation : removedOperations) {
			sb.append("operation " + umlOperation + " removed").append("\n");
		}
		Collections.sort(addedOperations);
		for(UMLOperation umlOperation : addedOperations) {
			sb.append("operation " + umlOperation + " added").append("\n");
		}
		Collections.sort(removedAttributes);
		for(UMLAttribute umlAttribute : removedAttributes) {
			sb.append("attribute " + umlAttribute + " removed").append("\n");
		}
		Collections.sort(addedAttributes);
		for(UMLAttribute umlAttribute : addedAttributes) {
			sb.append("attribute " + umlAttribute + " added").append("\n");
		}
		for(UMLOperationDiff operationDiff : operationDiffList) {
			sb.append(operationDiff);
		}
		for(UMLAttributeDiff attributeDiff : attributeDiffList) {
			sb.append(attributeDiff);
		}
		Collections.sort(operationBodyDiffList);
		for(UMLOperationBodyDiff operationBodyDiff : operationBodyDiffList) {
			sb.append(operationBodyDiff);
		}
		for(Entry<UMLOperation, UMLOperation> entry : changedOperationsNoParam.entrySet()){
			sb.append("operation " + entry.getKey() + " changed to " + entry.getValue()).append("\n");
		}
		for(Entry<UMLAttribute, UMLAttribute> entry : changedAttributes.entrySet()){
			sb.append("attribute " + entry.getKey() + " changed to " + entry.getValue()).append("\n");
		}
		return sb.toString();
	}

	public int compareTo(UMLClassDiff classDiff) {
		return this.className.compareTo(classDiff.className);
	}
}
